# modules
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
import ConfigParser
from base64 import b64encode, b64decode
from datetime import datetime
import errno
import exceptions
import getopt
import getpass
import httplib
import json
import logging.handlers
import os
import pwd
import signal
import shutil
import socket
import subprocess
import sys
import tempfile
import time
import traceback
import urllib
import uuid

try:
    import mysql.connector
except ImportError:
    # On CentOS 5 mysql.connector is in python 2.6 directory
    sys.path.insert(0, '/usr/lib/python2.6/site-packages')
    import mysql.connector

# global variables
host = "dispatcher.twindb.com"
proto = "http"
api_dir = ""
debug = False
debug_http = True
debug_gpg = False
init_config = "/etc/twindb.cfg"
ssh_private_key_file = "/root/.ssh/twindb.key"
ssh_public_key_file = "/root/.ssh/twindb.key.pub"
ssh_port = 4194
gpg_homedir = "/root/.gnupg/"
pid_file = "/var/run/twindb.pid"
check_period = 1
time_zone = "UTC"
api_email = "api@twindb.com"
server_id = ""
job_id = None
mysql_user = None
mysql_password = None
agent_version = "@@TWINDB_AGENT_VERSION@@"
api_pub_key = """
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1

mQINBFLAeJUBEADX+U107hrfNgcJQf9bBqZ6RLaYon4o2+mM0bMJwKg+tWeNlWWv
nsNknebq7ZoCz1YhNBYvYN9M4e3IM3qh+If/pMT+ZMad1bfVpIRk26VtCy19TSMt
KBfhagT7lZ9sVq6WKr0zQVQiGU2Ra7xmwdHnS2qfhppN77of6S+QywidwWME6gBG
WA1kBXvm2LOyIborZpiOjX8HyeNvTxlgEz9+PAkuANxRMrV22g+bGovyIN0g89zT
JqntFNhVJS0b+lG1TNYr+9lXKM3ZUY/PXi85akIgsC4c9odXzKjWV/DBi96IkfMk
NSolQ6Vpabgcl04GU/R5s8V1bBawReQ9kN4HaECkGvdgH7yBuIRcrE8LotP+Z9ax
qh9egTom3ezYtfKhYVQ2j5tEKF3FnCYXLsnT77ZyT4yyFwRvH+OgBM4um66dwCUp
qIqAzppyzOGrtV/Rpb210sUUV/xb+vUAaWrzFzyXVOxSRh/u1lxBEsSeHiv9OuJ+
TubvQVUV3U5/UdPjSToFuy6aQm7WOIChQY/qbdKzL55kLgpnYJunyBZcxH1kINQj
xN0FBrwld5Gn8sC8GUyC2gWfFTyfs3BE20XmX4c4zfkziraiA196j918nYF76HjT
KBRIRPxRMmqiT4Zs0P3ibhA5aU7mCr0vVEd/QEq0qhrzdLrVU5GGGcqi8QARAQAB
tCtUd2luREIgKFR3aW5EQiBBUEkgUG9ydGFsKSA8YXBpQHR3aW5kYi5jb20+iQI4
BBMBAgAiBQJSwHiVAhsDBgsJCAcDAgYVCAIJCgsEFgIDAQIeAQIXgAAKCRAjY6AO
qK2YiZVvEADNOf8fzubuLqHqFXtgK1D0WYJJVmllD1VV3jdOfayHyY76JcfGktiE
6EOzjTtG1sRvXJ2yU8KYJqCE4MaTwAugZKOKpH5fPoP3/Z35nXxGC5ZkbwdWynk0
BFdEz3cgOGyOU8XCTQ4/ctVEKc+jAnIZvoAoDKgK/9YRraM8Hbcfdn8XU7vAhvsu
aEUWLv6PIl+vUTnhFWSCGYLjeBBd/VnvRP9G4Krn80ZctEMOkzOqyH1J9vky1DzX
aXewW9K+AiLnUcIoWVMN5K3nPYu5YqBwzQMfvj6NR46rIj1Iv6MYzsrEPsYaWp/n
GQvHlLkvDr6jhYIbjF8VO3qWZ4PR4hgXTtKsKvu6NN5sltca4zSjqENvKmQHnnJc
xh+9s8Wir1NbtppNCf5fbtW7esapfWBouQh+ftjKrQeTaG97wLT8Ke2hgaFa5x86
0oEtWI/6nzLD/VcHikkCJrF4aimF9MADJJTXsoqN5N+CiYTLekCLu6ijHL1IAlNE
60jGJv/cZeOlXmv77bJdo0T/Haa1paxIPUvyBQk8cZaL8POaMAcoCVSX3aqlrf76
SKtYHch++b/LFpGmzEYBJMqlmw8BRuFH5qxKWC3e2F2cgGZWtrWYk0fnAg/mPZS+
mAg10IW7A5TU26RELEON0jeCX8cNwg7yDOWGt/RmRaTrD5j61P7afLkCDQRSwHiV
ARAA46Oa+4zSdPdjiR0nMLTuRClMTCowfUF20oVWt+UduGgrXZKqcAudH84r1Gyu
9SCc/RgOwEFEHAXljuNGl6WrXpw+6Xs6ptN0V1QJIQ5mVc82Nj8Or5WvaojiOzmr
0sI1Hpj2gyrdmbdEwC7vzpA5EuBbrVyh51/xJA9+3N66rJAbpGtknISrFZ1EXsRH
MjRaaPT/xaVvVJgsp7wIG357kbuXC5MH4bcaGQgIPfqk8VPx96UugzMgSDC8Bs+t
Oj5+U0cRFxQFu99LWm6wLUIvpR+8vo36IHGVuVzhjcguzj3WQSCieq/QBbovRXQX
3UZSCfUu8hUscTTugaq8iP504OOJEWzvCcpbGlHYGMGDMRzH7FraKlz1D1h7FNtd
ytJryyFva3KcjJbTGyLV4kDkWUwk8Ysu/O+2LS2I7NNA7se/3exsJq/k/McatDpQ
JScZpKLddi8VsjBgSAvpmgtIWL9yuJql3evIC9SGK4Jk5mIzV6edZDh8e7EbwjMM
zhw58UXbqtGAK2I7SDo0FK4sSUBW76Tv+0+vOnq6Jlb+hjieQrAHaMk5tBQtlCNI
hdE8D24ZwLPyYoHCjUluszncYwAvdub1F0OEs7YMfsDG57jfj+5CTrWK9V3q/wMW
+68rlIf2Xf3LgQ+tBi4z7RcgAhjpbr+5wHQZkN3fsPX3rU8AEQEAAYkCHwQYAQIA
CQUCUsB4lQIbDAAKCRAjY6AOqK2YiZZbD/0SS+Su4bJ8Gj55ekcSudof16IAXYhH
Iv2knjPRAQV4GnJV8zaYSrSxz6ZAGrX9s3emO5t94Sp23xtCyhIzQLNHOfu2kvg9
9cX4QIJwDKU6YvDwmG/onuZW00uzMs8qJyFQcQrgbnG6q6c55INypI9UT5wRDhVf
r8zAOpUTgG90oghRr7onH/v6nbzqPlO4u0b4Mh933XuxGfROb4MZZX8amCcOMiXG
zvBqQ3AyEhFXBrRst7KVWggwmgyCzRwft6FGmRcoyyujvjszyutha7jLikl3Rcsw
JOp4uqXQ7TcM5frLTCa1TDLCsuk6C6EgD96wHjsK+LM/z4/giweh+toKEnqF7UJD
I4k4SIeEMeKl22VuBHjz1m27l+HL6a15X/A65a2NblauUZJV/jAbm5Qkapxc1vjs
oyCVh3aPH+BPIfWGVE27EWMupRvLGg0ufVxN/zK0En7osjIkwtdGSRPDJ3sDWuq9
Uv1H8Ezizp9ji8I6XIZSICiUfdrmur4bGGCp5CanLcNjIPA1uIWCPaXr/SpHWuVA
4bgIYkEK171jZS60GqFwLWPNiGCmXwyQqykVX1/HqiQBApNPYNU6/z5k6pW+zQqt
YbniyuJvMcKt2szQS07cqbgh7yevaSHw/L6KX1Ih6CFmL2ZG1oo5zIsJBYJeEQYL
Q7n4+TygXY9zVA==
=62IM
-----END PGP PUBLIC KEY BLOCK-----
"""


class RlogHandler(logging.Handler):
    """
    Logging handler that logs to remote TwiDB dispatcher
    """
    log_flag = True

    def __init__(self):
        logging.Handler.__init__(self)
        self.log_flag = True

    def emit(self, record):
        request = {
            "type": "log",
            "params": {}
        }
        if job_id:
            request["params"]["job_id"] = job_id
        request["params"]["msg"] = record.getMessage()
        if self.log_flag and not debug:
            self.log_flag = False
            get_response(request)
            self.log_flag = True


class JobError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


class JobTooSoonError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


def usage():
    print("""Usage:
twindb --start | --stop | --register <registration code> [-g] [-i interval]
       [-u user] [-p password]

    --start                         Start TwinDB agent.
    --stop                          Stop TwinDB agent.

    --register <code>               Register this server in TwinDB.
                                    <code> is string like 7b90ae21ac642f2f8fc0a285c4789147
                                    Get your code on https://console.twindb.com/?get_code

    --unregister [--delete-backups] Unregister this server from TwinDB.
    --backup                        Start backup from this server immediately.
    --is-registered                 Checks whether the server is registered and prints YES or NO.
    -g                              Print debug message.
    -i interval                     Check for a job every interval seconds.
                                    Default {check_period}.
                                    During registration TwinDB agent needs to get some information from MySQL,
                                    (master host etc)
                                    This is credentials to connect to local MySQL instance
    -u | --user username            MySQL user name
    -p | --password password        MySQL password""".format(check_period=check_period))
    return


def get_mysql_connection(user=None, passwd=None):
    """
    Returns connection to the local MySQL instance.
    If user is passed as an argument it'll be used to connect,
    otherwise the second choice will be to use mysql_user.
    If neither user names are set the function will try to use either of MySQL option files
    (/etc/my.cnf, /etc/mysql/my.cnf, or /root/.my.cnf). If the option files don't exist
    it'll try to connect as root w/o password.
    """
    try:
        unix_socket = get_unix_socket()
        if not user:
            if mysql_user or mysql_password:
                logger.debug('Using MySQL user specified in the command line')
                if mysql_user:
                    user = mysql_user
                else:
                    user = getpass.getuser()
                passwd = mysql_password
            else:
                for options_file in ["/etc/my.cnf", "/etc/mysql/my.cnf", "/usr/etc/my.cnf",
                                     "/root/.my.cnf", "/root/.mylogin.cnf"]:
                    if os.path.exists(options_file):
                        try:
                            config = ConfigParser.ConfigParser()
                            config.read(options_file)
                            for section in ["client", "twindb"]:
                                if config.has_section(section):
                                    if config.has_option(section, "user"):
                                        user = config.get(section, "user")
                                    if config.has_option(section, "password"):
                                        passwd = config.get(section, "user")
                        except ConfigParser.ParsingError as err:
                            logger.debug(err)
                            logger.debug("Ignoring options file %s" % options_file)
                            pass
                # If user isn't set by the function argument, global mysql_user
                # or MySQL options file connect as unix user w/ empty password
                if not user:
                    user = pwd.getpwuid(os.getuid()).pw_name
                    passwd = ""
                    logger.debug("Connecting to MySQL as unix user %s" % user)
        conn = mysql.connector.connect(user=user, passwd=passwd, unix_socket=unix_socket)
        logger.debug("Connected to MySQL as %s@localhost " % conn.user)
    except mysql.connector.Error as err:
        logger.error("Can not connect to local MySQL server")
        logger.error("MySQL said: %s" % err.msg)
        return None
    return conn


def is_gpg_key_installed(email, key_type="public"):
    """
    Checks if TwinDB API public key is installed
    Returns
    True    - if the key is installed
    False   - if not
    """
    result = False
    if not email:
        logger.error("Can not check public key of an empty email")
        return False
    if debug_gpg:
        logger.debug("Checking if public key of %s is installed" % email)
    gpg_cmd = ["gpg", "--homedir", gpg_homedir]
    try:
        if key_type == "public":
            gpg_cmd.append("-k")
        else:
            gpg_cmd.append("-K")
        gpg_cmd.append(email)
        if debug_gpg:
            logger.debug(gpg_cmd)
        devnull = open("/dev/null", "w")
        ret = subprocess.call(gpg_cmd, stdout=devnull, stderr=devnull)
        devnull.close()
        if ret != 0:
            if debug_gpg:
                logger.debug("GPG returned %d" % ret)
                logger.debug(key_type + " key " + email + " is NOT installed")
            result = False
        else:
            if debug_gpg:
                logger.debug(key_type + " key is already installed")
            result = True
    except OSError as err:
        logger.error("Couldn't run command %r. %s" % (gpg_cmd, err))
        exit_on_error("Couldn't get %s key of %s." % (key_type, email))
    return result


def install_api_pub_key():
    """
    Installs TwinDB public key
    Returns
      True    - if the key is successfully installed
    Exits if the key wasn't installed
    """
    logger.info("Installing twindb public key")
    gpg_cmd = ["gpg", "--homedir", gpg_homedir, "--import"]
    try:
        p = subprocess.Popen(gpg_cmd, stdin=subprocess.PIPE)
        p.communicate(api_pub_key)
    except OSError as err:
        logger.error("Couldn't run command %r. %s" % (gpg_cmd, err))
        exit_on_error("Couldn't install TwinDB public key")
    logger.info("Twindb public key successfully installed")
    return True


def check_gpg():
    """
    Checks if GPG environment is good to start TwinDB agent
    Installs TwinDB public key if necessary
    Returns
      True    - if the GPG environment good to proceed
    Exits if GPG wasn't configured correctly
    """
    if debug_gpg:
        logger.debug("Checking if GPG config is initialized")
    if os.path.exists(gpg_homedir):
        if not is_gpg_key_installed(api_email):
            install_api_pub_key()
        else:
            if debug_gpg:
                logger.debug("Twindb public key is already installed")
    else:
        logger.info("Looks like GPG never ran. Initializing GPG configuration")
        try:
            os.mkdir(gpg_homedir, 0700)
            install_api_pub_key()
        except OSError as err:
            logger.error("Failed to create directory %s. %s" % (gpg_homedir, err))
            exit_on_error("Couldn't create directory " + gpg_homedir)
    if not server_id:
        exit_on_error("Server id is neither read from config nor generated")
    email = "%s@twindb.com" % server_id
    if not (is_gpg_key_installed(email) and is_gpg_key_installed(email, "private")):
        gen_entropy()
        gen_gpg_keypair("%s@twindb.com" % server_id)
    if debug_gpg:
        logger.debug("GPG config is OK")
    return True


def check_env():
    """
    Checks the environment if it's OK to start TwinDB agent
    Returns
      True    - if the environment is OK
    Exits if the environment is not OK
    """
    logger.debug("Checking environment")
    if os.getuid() != 0:
        exit_on_error("TwinDB agent must be run by root")
    check_gpg()
    logger.debug("Environment is OK")
    return True


def gen_entropy():
    """
    Checks how much entropy is available in the system
    If not enough, does some disk activity to generate more
    """
    # Do nothing until I find good way to generate entropy
    return


def h_size(num, decimals=2):
    """
    Formats bytes count to human readable form
    Inputs:
      num         - bytes count
      decimals    - number of digits to save after point (Default 2)
    Returns
      human-readable string like "20.33 MB"
    """
    fmt = "%3." + str(decimals) + "f %s"
    for x in ['bytes', 'kB', 'MB', 'GB', 'TB', 'PB']:
        if num < 1024.0:
            return fmt % (num, x)
        num /= 1024.0


def gen_gpg_keypair(email):
    """
    Generates GPG private and public keys for a given recipient
    Inputs
      email   - recipient
    Returns
      True on success or exits
    """
    if not email:
        exit_on_error("Can not generate GPG keypair for an empty email")
    gpg_cmd = ["gpg", "--homedir", gpg_homedir, "--batch", "--gen-key"]
    try:
        logger.info("The agent needs to generate cryptographically strong keys.")
        logger.info("Generating GPG keys pair for %s" % email)
        logger.info("It may take really, really long time. Please be patient.")
        gen_entropy()
        gpg_script = """
%%echo Generating a standard key
Key-Type: RSA
Key-Length: 2048
Subkey-Type: RSA
Subkey-Length: 2048
Name-Real: Backup Server id %s
Name-Comment: No passphrase
Name-Email: %s
Expire-Date: 0
%%commit
%%echo done
""" % (server_id, email)

        p = subprocess.Popen(gpg_cmd, stdin=subprocess.PIPE)
        p.communicate(gpg_script)
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (gpg_cmd, err))
        exit_on_error("Failed to generate GPG keys pair")
    return True


def encrypt(msg):
    """
    Encrypts message with TwinDB public key
    If server_id is non-zero (which means the server is registered)
    signs the message with the server's private key
    :param msg: string to encrypt
    :return:  64-base encoded and encrypted message or None if error happens.
    To read the encrypted message - decrypt and 64-base decode
    """
    if not server_id:
        exit_on_error("Trying to encrypt message while server_id is empty")
    server_email = "%s@twindb.com" % server_id
    if not is_gpg_key_installed(api_email):
        logger.debug("Public key of %s is not installed." % api_email)
        logger.debug("Will not encrypt message")
        return None
    if not is_gpg_key_installed(server_email):
        logger.debug("Public key of %s is not installed." % server_email)
        logger.debug("Will not encrypt message")
        return None
    enc_cmd = ["gpg", "--homedir", gpg_homedir, "-r", api_email, "--batch", "--trust-model", "always", "--armor",
               "--sign", "--local-user", server_email, "--encrypt"]
    cout = "No output"
    cerr = "No output"
    try:
        if debug_gpg:
            logger.debug("Encrypting message:")
            logger.debug(msg)
            logger.debug("Encryptor command: ")
            logger.debug(enc_cmd)
        p = subprocess.Popen(enc_cmd,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        cout, cerr = p.communicate(msg)
        if p.returncode != 0:
            raise OSError(p.returncode)
        ct = cout
        ct_64 = b64encode(ct)
        if debug_gpg:
            logger.debug("Encrypted message: " + ct_64)
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (enc_cmd, err))
        logger.error("Failed to encrypt message: " + msg)
        logger.error("STDOUT: " + cout)
        logger.error("STDERR: " + cerr)
        return None
    return ct_64


def decrypt(msg_64):
    """
    Decrypts message with local private key
    :param msg_64: 64-base encoded and encrypted message. Before encryption the message was 64-base encoded
    :return: Plain text message or None if error happens
    """
    if not msg_64:
        logger.error("Will not decrypt empty message")
        return None
    cout = "No output"
    cerr = "No output"
    gpg_cmd = ["gpg", "--homedir", gpg_homedir, "-d", "-q"]
    try:
        if debug_gpg:
            logger.debug("Decrypting message:")
            logger.debug(msg_64)
            logger.debug("Decryptor command:")
            logger.debug(gpg_cmd)
        p = subprocess.Popen(gpg_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        msg = b64decode(msg_64)
        cout, cerr = p.communicate(msg)
        if p.returncode != 0:
            raise OSError(p.returncode)
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (gpg_cmd, err))
        logger.error("Failed to decrypt message: " + msg_64)
        logger.error("STDOUT: " + cout)
        logger.error("STDERR: " + cerr)
        return None
    if debug_gpg:
        logger.debug("Decrypted message:")
        logger.debug(cout)
    return cout


def get_response(request):
    """
    Sends HTTP POST request to TwinDB dispatcher
    It converts python data structure in "data" variable into JSON string,
    then encrypts it and then sends as variable "data" in HTTP request
    Inputs
        uri     - URI to send the request
        data    - Data structure with variables
    Returns
        String with body of HTTP response
        None    - if error happened or empty response
    """
    uri = "api.php"
    response_body = None
    logger.debug("Enter get_response(uri=" + uri + ")")
    if proto == "http":
        conn = httplib.HTTPConnection(host)
    elif proto == "https":
        conn = httplib.HTTPSConnection(host)
    else:
        logger.error("Unsupported protocol " + proto)
        return None
    url = proto + "://" + host + "/" + api_dir + "/" + uri
    http_response = "Empty response"
    try:
        conn.connect()
        logger.debug("Sending to " + host + ": %s" % json.dumps(request, indent=4, sort_keys=True))
        data_json = json.dumps(request)
        data_json_enc = encrypt(data_json)
        data_json_enc_urlenc = urllib.urlencode({'data': data_json_enc})
        conn.putrequest('POST', "/" + api_dir + "/" + uri)
        headers = dict()
        headers['Content-Length'] = "%d" % (len(data_json_enc_urlenc))
        headers['Content-Type'] = 'application/x-www-form-urlencoded'
        for k in headers:
            conn.putheader(k, headers[k])
        conn.endheaders()
        conn.send(data_json_enc_urlenc)
        http_response = conn.getresponse()

        if http_response.status == 200:
            response_body = http_response.read()
            logger.debug("Response body: '%s'" % response_body)
            if len(response_body) == 0:
                return None
            url = "%(proto)s://%(host)s/%(api_dir)s/%(uri)s" % {
                "proto": proto,
                "host": host,
                "api_dir": api_dir,
                "uri": uri
            }
            d = json.JSONDecoder()
            try:
                json.loads(response_body)
            except ValueError:
                msg = response_body
            else:
                msg = json.dumps(d.decode(response_body), indent=4, sort_keys=True)
            logger.debug("Response from %(url)s : '%(resp)s'" % {
                "url": url,
                "resp": msg
            })
        else:
            logger.error("HTTP error %d %s" % (http_response.status, http_response.reason))
    except socket.error as err:
        logger.error("Exception while making request " + url)
        logger.error(err)
        logger.error("TwinDB dispatcher is unreachable")
        return None
    except exceptions.KeyError as err:
        logger.error("Failed to decode response from server: %s" % http_response)
        logger.error("Could not find key %s" % err)
        return None
    finally:
        conn.close()
    return response_body


def sanitize_config(config):
    """
    Replaces password from config with ********
    :param config: python dictionary with backup config
    :return: Sanitized config
    """
    if not config:
        return None
    sanitized_config = dict(config)
    if "mysql_password" in config:
        sanitized_config["mysql_password"] = "********"
    else:
        logger.debug("Given config %r doesn't contain password" % config)
    return sanitized_config


def get_config():
    """
    Gets backup config from TwinDB dispatcher
    :return: Backup config or None if error happened
    """
    logger.debug("Getting config for server_id = %s" % server_id)
    response_body = "Empty response"
    config = None
    try:
        data = {
            "type": "get_config",
            "params": {
                "server_id": server_id
            }
        }
        response_body = get_response(data)
        if not response_body:
            return None
        d = json.JSONDecoder()
        response_body_decoded = d.decode(response_body)
        if response_body_decoded:
            msg_decrypted = decrypt(response_body_decoded["response"])
            msg_pt = d.decode(msg_decrypted)
            config = msg_pt["data"]
            logger.info("Got config:\n%s" % json.dumps(sanitize_config(config), indent=4, sort_keys=True))
            if msg_pt["error"]:
                logger.error(msg_pt["error"])
    except exceptions.KeyError as err:
        logger.error("Failed to decode %s" % response_body)
        logger.error(err)
        return None
    return config


def report_sss(config):
    """
    Reports slave status to TwinDB dispatcher
    :param config: server config received from the dispatcher
    :return: nothing
    """
    logger.debug("Reporting SHOW SLAVE STATUS for server_id = %s" % server_id)
    ss = get_slave_status(config["mysql_user"], config["mysql_password"])
    try:
        data = {
            "type": "report_sss",
            "params": {
                "server_id": server_id,
                "mysql_server_id": ss["mysql_server_id"],
                "mysql_master_server_id": ss["mysql_master_server_id"],
                "mysql_master_host": ss["mysql_master_host"],
                "mysql_seconds_behind_master": ss["mysql_seconds_behind_master"],
                "mysql_slave_io_running": ss["mysql_slave_io_running"],
                "mysql_slave_sql_running": ss["mysql_slave_sql_running"],
            }
        }
        get_response(data)
    except TypeError as err:
        logger.error("Could not report replication status to the dispatcher")
        logger.error(err)
    return


def report_agent_privileges(config):
    """
    Reports what privileges are given to the agent
    :param config: server config received from the dispatcher
    :return: nothing
    """
    logger.debug("Reporting agent privileges for server_id = %s" % server_id)
    try:
        con = get_mysql_connection(user=config["mysql_user"], passwd=config["mysql_password"])
        cursor = con.cursor()
        query = "SELECT PRIVILEGE_TYPE FROM information_schema.USER_PRIVILEGES"
        logger.debug("Sending query : %s" % query)
        cursor.execute(query)

        privileges = {
            "Reload_priv": "N",
            "Lock_tables_priv": "N",
            "Repl_client_priv": "N",
            "Super_priv": "N",
            "Create_tablespace_priv": "N"
        }
        for (priv,) in cursor:
            if priv == "RELOAD":
                privileges["Reload_priv"] = "Y"
            elif priv == "LOCK TABLES":
                privileges["Lock_tables_priv"] = "Y"
            elif priv == "REPLICATION CLIENT":
                privileges["Repl_client_priv"] = "Y"
            elif priv == "SUPER":
                privileges["Super_priv"] = "Y"
            elif priv == "CREATE TABLESPACE":
                privileges["Create_tablespace_priv"] = "Y"
        data = {
            "type": "report_agent_privileges",
            "params": {
                "Reload_priv": privileges["Reload_priv"],
                "Lock_tables_priv": privileges["Lock_tables_priv"],
                "Repl_client_priv": privileges["Repl_client_priv"],
                "Super_priv": privileges["Super_priv"],
                "Create_tablespace_priv": privileges["Create_tablespace_priv"]
            }
        }
        get_response(data)
    except AttributeError as err:
        logger.error("Could not report agent permissions to the dispatcher")
        logger.error(err)
    return


def get_job():
    """
    Gets job order from TwinDB dispatcher
    :return: Job order in python dictionary or None if error happened
    """
    job = None
    logger.debug("Getting job for server_id = %s" % server_id)
    try:
        d = json.JSONDecoder()
        data = {
            "type": "get_job",
            "params": {}
        }
        response_body = get_response(data)
        if not response_body:
            raise JobError("Empty response from dispatcher")
        response_body_decoded = d.decode(response_body)
        if "response" not in response_body_decoded:
            raise JobError("There is no 'response' key in the response from dispatcher")
        if "success" not in response_body_decoded:
            raise JobError("There is no 'success' key in the response from dispatcher")
        msg_enc = response_body_decoded["response"]
        if response_body_decoded["success"]:
            job_json = decrypt(msg_enc)
            logger.debug("job_json = %s" % job_json)
            if "data" not in job_json:
                raise JobError("There is no 'data' in decrypted response %s" % job_json)
            job = d.decode(job_json)["data"]
            if job and "params" in job and job["params"]:
                job["params"] = d.decode(job["params"])
                logger.info("Got job:\n%s" % json.dumps(job, indent=4, sort_keys=True))
        else:
            logger.error("Couldn't get job")
            job_json = decrypt(msg_enc)
            logger.error("Server said: %s" % d.decode(job_json)["error"])
    except exceptions.TypeError as err:
        logger.error("Failed to get a job: %s" % err)
        return None
    except JobError as err:
        logger.error(err)
        return None
    return job


def is_registered():
    """
    Checks whether the server is registered or not
    :return: True if registered, False if not so much
    """
    logger.debug("Getting registration status for server_id = %s" % server_id)

    twindb_email = "%s@twindb.com" % server_id
    logger.debug("Reading GPG public key of %s." % twindb_email)

    enc_public_key = None
    # Reading the GPG key
    gpg_cmd = ["gpg", "--homedir", gpg_homedir, "--armor", "--export", twindb_email]
    try:
        p = subprocess.Popen(gpg_cmd, stdout=subprocess.PIPE)
        enc_public_key = p.communicate()[0]
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (gpg_cmd, err))
        exit_on_error("Failed to export GPG keys of %s from %s." % (twindb_email, gpg_homedir))

    # Call the TwinDB api to check for server registration
    response_body = None
    try:
        data = {
            "type": "is_registered",
            "params": {
                "server_id": server_id,
                "enc_public_key": enc_public_key
            }
        }
        response_body = get_response(data)
        if not response_body:
            return False
        json_decoder = json.JSONDecoder()
        response_body_decoded = json_decoder.decode(response_body)
        if response_body_decoded:
            if "response" in response_body_decoded:
                msg_decrypted = decrypt(response_body_decoded["response"])
                if msg_decrypted is None:
                    logger.debug("No valid response from dispatcher. Consider agent unregistered")
                    return False
                msg_pt = json_decoder.decode(msg_decrypted)
                registration_status = msg_pt["data"]
                logger.debug("Got registration status:\n%s" % json.dumps(registration_status, indent=4, sort_keys=True))
                if msg_pt["error"]:
                    logger.error(msg_pt["error"])
                    exit_on_error(msg_pt["error"])
                return registration_status["registered"]
            else:
                logger.debug("No valid response from dispatcher. Consider agent unregistered")
                return False
    except ValueError:
        logger.debug("No valid JSON response from dispatcher. Consider agent unregistered")
        return False
    except exceptions.KeyError as err:
        exit_on_error("Failed to decode response from dispatcher: %s. %s" % (response_body, err))
    return False


def log_job_notify(params):
    """
    Notifies a job event to TwinDB dispatcher
    :param params: { event: "start_job", job_id: job_id } or
             { event: "stop_job", job_id: job_id, ret_code: ret }
    :return: True of False if error happened
    """
    logger.info("Sending event notification %s" % params["event"])
    data = {
        "type": "notify",
        "params": params
    }
    response_body = get_response(data)
    d = json.JSONDecoder()
    response_body_decoded = d.decode(response_body)
    if response_body_decoded["success"]:
        logger.debug("Dispatcher acknowledged job_id = %d notification" % job_id)
        result = True
    else:
        logger.error("Dispatcher didn't acknowledge job_id = %d notification" % job_id)
        result = False
    return result


def record_backup(name, volume_id, size, lsn=None, ancestor=0):
    """
    Saves details about backup copy in TwinDB dispatcher
    :param name: name of backup
    :param volume_id: id of a volume where the backup copy is saved
    :param size: size of the backup in bytes
    :param lsn: last LSN if it was incremental backup
    :param ancestor: Ansector of the backup (not used now)
    :return: JSON string with status of the request i.e. { "success": True } or None if error happened
    """
    logger.info("Saving information about backup:")
    logger.info("File name : %s" % name)
    logger.info("Volume id : %d" % int(volume_id))
    logger.info("Size      : %d (%s)" % (int(size), h_size(size)))
    logger.info("Ancestor  : %d" % int(ancestor))
    data = {
        "type": "update_backup_data",
        "params": {
            "job_id": job_id,
            "name": name,
            "volume_id": volume_id,
            "size": size,
            "lsn": lsn,
            "ancestor": ancestor
        }
    }
    logger.debug("Saving a record %s" % data)
    response = get_response(data)
    if response:
        jd = json.JSONDecoder()
        r = jd.decode(response)
        logger.debug(r)
        if r["success"]:
            logger.info("Saved backup copy details")
            return True
        else:
            logger.error("Failed to save backup copy details: "
                         + jd.decode(decrypt(r["response"]))["error"])
            return False
    else:
        logger.error("Empty response from server")
        return False


def read_config():
    """
    Reads config from /etc/twindb.cfg and sets server_id variable
    :return: True if config was successfully read
    Exits if error happened
    """
    global server_id
    try:
        if os.path.exists(init_config):
            ns = {}
            execfile(init_config, ns)
            if "server_id" in ns:
                server_id = ns["server_id"]
        else:
            print("Config %s doesn't exist" % init_config, file=sys.stderr)
            return False
        if not server_id:
            exit_on_error("Config %s doesn't set server_id" % init_config)
    except IOError as err:
        exit_on_error("Failed to read config file %s. %s" % (init_config, err))
    return True


def save_config():
    """
    Saves server_id variable in file init_config (/etc/twindb.cfg)
    :return: True    - if config was successfully saved
    Exits if error happened
    """
    if not server_id:
        exit_on_error("Can not save agent config file because server_id is empty")
    try:
        f = open(init_config, "w")
        f.write("server_id='%s'\n" % server_id)
        f.write("host='%s'\n" % host)
        f.write("api_pub_key=\"\"\"%s\"\"\"\n" % api_pub_key)
        f.close()
    except IOError as err:
        exit_on_error("Failed to save new config in %s. %s" % (init_config, err))
    return True


def get_slave_status(user=None, passwd=None):
    """
    Reads SHOW SLAVE STATUS from the local server
    :param user: user to connect to MySQL
    :param passwd: password to connect to MySQL
    :return: dictionary with SHOW SLAVE STATUS result
    """
    conn = get_mysql_connection(user, passwd)
    if conn:
        cursor = conn.cursor(dictionary=True)
    else:
        return None
    result = {
        "mysql_server_id": None,
        "mysql_master_server_id": None,
        "mysql_master_host": None,
        "mysql_seconds_behind_master": None,
        "mysql_slave_io_running": None,
        "mysql_slave_sql_running": None
    }
    try:
        cursor.execute("SHOW SLAVE STATUS")
        for row in cursor:
            result["mysql_master_server_id"] = row["Master_Server_Id"]
            result["mysql_master_host"] = row["Master_Host"]
            result["mysql_seconds_behind_master"] = row["Seconds_Behind_Master"]
            result["mysql_slave_io_running"] = row["Slave_IO_Running"]
            result["mysql_slave_sql_running"] = row["Slave_SQL_Running"]
    except mysql.connector.Error as err:
        logger.error("Could get SHOW SLAVE STATUS")
        logger.error("MySQL Error: " % err)
        return None
    try:
        cursor.execute("SELECT @@server_id AS server_id")
        for row in cursor:
            result["mysql_server_id"] = row["server_id"]
        cursor.close()
        conn.close()
    except mysql.connector.Error as err:
        logger.error("Could not read server_id")
        logger.error("MySQL Error: " % err)
        return None
    return result


def has_mysql_access(username=None, password=None, grant_capability=True):
    """
    Reports if a user has all required MySQL privileges
    :param username: MySQL user
    :param password: MySQL password
    :param grant_capability: TODO to add description
    :return: a pair of a boolean that tells whether MySQL user has all required privileges
    and a list of missing privileges
    """
    has_required_grants = False

    # list of missing privileges
    missing_privileges = []
    required_privileges = ['RELOAD', 'SUPER', 'LOCK TABLES', 'REPLICATION CLIENT', 'CREATE TABLESPACE']

    try:
        conn = get_mysql_connection(username, password)

        if isinstance(conn, mysql.connector.MySQLConnection):
            cursor = conn.cursor(dictionary=True)
        else:
            missing_privileges = required_privileges
            return has_required_grants, missing_privileges

        # Fetch the current user and matching host part as it could either be
        # connecting using localhost or using '%'
        cursor.execute("SELECT CURRENT_USER() as curr_user")
        row = cursor.fetchone()
        username, hostname = row['curr_user'].split('@')

        quoted_privileges = ','.join("'%s'" % item for item in required_privileges)

        sql = ("SELECT privilege_type, is_grantable FROM information_schema.user_privileges "
               "WHERE grantee=\"'%s'@'%s'\" AND privilege_type IN (%s)" % (username, hostname, quoted_privileges))
        cursor.execute(sql)

        user_privileges = []
        grantable_privileges = []
        for row in cursor:
            user_privileges.append(row[u'privilege_type'])
            if row[u'is_grantable'] == 'YES':
                grantable_privileges.append(row[u'privilege_type'])

        # Check that the user has all the required grants
        has_required_grants = True
        for privilege in required_privileges:
            if privilege in user_privileges:
                # If the user should also be able to grant the privilege then we check for the grant capability too
                # We consider the privilege not available if its not grantable in such a case
                if grant_capability and privilege not in grantable_privileges:
                    has_required_grants = False
            else:
                has_required_grants = False
                missing_privileges.append(privilege)

        if len(missing_privileges) < 1 and grant_capability:
            required_privileges = ['INSERT', 'UPDATE']
            quoted_privileges = ','.join("'%s'" % item for item in required_privileges)

            # If the user should be able to grant privileges too, such as the user that is used to create the
            # twindb_agent user then, insert and update privileges are needed on mysql.*
            sql = ("SELECT privilege_type FROM information_schema.schema_privileges "
                   "WHERE grantee=\"'%s'@'%s'\" AND table_schema = 'mysql' AND privilege_type IN (%s)"
                   "UNION "
                   "SELECT privilege_type FROM information_schema.user_privileges "
                   "WHERE grantee=\"'%s'@'%s'\" AND privilege_type IN (%s)"
                   % (username, hostname, quoted_privileges, username, hostname, quoted_privileges))
            cursor.execute(sql)

            user_privileges = []
            for row in cursor:
                user_privileges.append(row[u'privilege_type'])

            for privilege in required_privileges:
                if privilege not in user_privileges:
                    has_required_grants = False
                    break

        cursor.close()
        conn.close()
    except mysql.connector.Error as err:
        logger.error("Could not read the grants information from MySQL")
        logger.error("MySQL Error: %s" % err)
    return has_required_grants, missing_privileges


def action_handler_register(code):
    """
    Registers this server in TwinDB dispatcher
    Inputs
      code    - string with secret registration code
    Returns
      True    - if server was successfully registered
    Exits if error happened
    """
    # Check that the agent can connect to local MySQL
    conn = get_mysql_connection()
    if conn:
        conn.close()
    else:
        if not mysql_user or mysql_password:
            logger.info("Try to call the agent with -u and -p options to specify MySQL user and password")
            logger.info("Another option is to specify MySQL user and password in /root/.my.cnf")
        sys.exit(2)
    # Check early to see that the MySQL user passed to the agent has enough
    # privileges to create a separate MySQL user needed by TwinDB
    mysql_access_available, missing_mysql_privileges = has_mysql_access()
    if not mysql_access_available:
        logger.error("The MySQL user %s does not have enough privileges" % mysql_user)
        if missing_mysql_privileges:
            logger.error("Following privileges are missing: %s" % ','.join(missing_mysql_privileges))
        return False

    logger.info("Registering TwinDB agent with code %s" % code)
    name = os.uname()[1].strip()  # un[1] is a hostname

    twindb_email = "%s@twindb.com" % server_id

    # Read GPG public key
    enc_public_key = None
    cmd = ["gpg", "--homedir", gpg_homedir, "--armor", "--export", twindb_email]
    try:
        logger.debug("Reading GPG public key of %s." % twindb_email)
        p1 = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        enc_public_key = p1.communicate()[0]
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (cmd, err))
        exit_on_error("Failed to export GPG keys of %s from %s." % (twindb_email, gpg_homedir))

    if not os.path.isfile(ssh_private_key_file):
        try:
            logger.info("Generating SSH keys pair.")
            subprocess.call(["ssh-keygen", "-N", "", "-f", ssh_private_key_file])
        except OSError as err:
            logger.error("Failed to run command %r. %s" % (cmd, err))
            exit_on_error("Failed to generate SSH keys.")
    ssh_public_key = None
    try:
        logger.info("Reading SSH public key from %s." % ssh_public_key_file)
        f = open(ssh_public_key_file, 'r')
        ssh_public_key = f.read()
        f.close()
    except IOError as err:
        logger.error("Failed to read from file %s. %s" % (ssh_public_key_file, err))
        exit_on_error("Failed to read SSH keys.")

    # Read local ip addresses
    cmd = "ip addr"
    cmd += "| grep -w inet"
    cmd += "| awk '{ print $2}'"
    cmd += "| awk -F/ '{ print $1}'"
    cout = None
    logger.debug("Getting list of local IP addresses")
    try:
        logger.debug("Running: %s" % cmd)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        cout, cerr = p.communicate()
        logger.debug("STDOUT: %s" % cout)
        logger.debug("STDERR: %s" % cerr)
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (cmd, err))
    local_ip = list()
    for row in cout.split("\n"):
        row = row.rstrip("\n")
        if row and row != "127.0.0.1":
            local_ip.append(row)
    ss = get_slave_status()
    data = {
        "type": "register",
        "params": {
            "reg_code": code,
            "name": name,
            "server_id": server_id,
            "enc_public_key": enc_public_key,
            "ssh_public_key": ssh_public_key,
            "mysql_server_id": ss["mysql_server_id"],
            "mysql_master_server_id": ss["mysql_master_server_id"],
            "mysql_master_host": ss["mysql_master_host"],
            "mysql_seconds_behind_master": ss["mysql_seconds_behind_master"],
            "mysql_slave_io_running": ss["mysql_slave_io_running"],
            "mysql_slave_sql_running": ss["mysql_slave_sql_running"],
            "local_ip": local_ip
        }
    }
    api_response = get_response(data)
    if api_response:
        json_decoder = json.JSONDecoder()
        response_decoded = json_decoder.decode(api_response)
        logger.debug(response_decoded)

        error_msg = "Unknown error"
        if response_decoded["success"]:
            logger.info("Received successful response to register an agent")
            create_agent_user()
        else:
            if "response" in response_decoded:
                msg_decrypted = decrypt(response_decoded["response"])
                if msg_decrypted is not None:
                    error_msg = json_decoder.decode(msg_decrypted)["error"]
            elif "errors" in response_decoded:
                error_msg = response_decoded["errors"]["msg"]

            exit_on_error("Failed to register the agent: %s" % error_msg)
    else:
        exit_on_error("Empty response from server")
    return True


def create_agent_user():
    """
    Creates local MySQL user for twindb agent
    """
    config = get_config()
    try:
        conn = get_mysql_connection()
        q = "GRANT RELOAD, LOCK TABLES, REPLICATION CLIENT, SUPER, CREATE TABLESPACE"
        q += " ON *.* TO %s@'localhost' IDENTIFIED BY %s"
        cursor = conn.cursor()
        cursor.execute(q, (config["mysql_user"], config["mysql_password"]))
    except mysql.connector.Error as err:
        logger.error("MySQL replied: %s" % err)
        exit_on_error("Failed to create MySQL user %s@localhost for TwinDB agent" % config["mysql_user"])
    logger.info("Created MySQL user %s@localhost for TwinDB agent" % config["mysql_user"])
    logger.info("Congratulations! The server is successfully registered in TwinDB.")
    logger.info("TwinDB will backup the server accordingly to the default config.")
    logger.info("You can change the schedule and retention policy on https://console.twindb.com/")
    return True


def action_handler_unregister(delete_backups):
    """
    Unregisters this server in TwinDB dispatcher
    Returns
      True    - if server was successfully unregistered
      False   - if error happened
    """
    data = {
        "type": "unregister",
        "params": {
            "delete_backups": delete_backups,
        }
    }
    logger.debug("Unregistration request:")
    logger.debug(data)
    response = get_response(data)
    if response:
        jd = json.JSONDecoder()
        r = jd.decode(response)
        logger.debug(r)
        if r["success"]:
            logger.info("The server is successfully unregistered")
            return True
        else:
            exit_on_error("Failed to unregister the agent: " + jd.decode(decrypt(r["response"]))["error"])
    else:
        exit_on_error("Empty response from server")
    return False


def action_handler_backup():
    """
    Starts immediate backup job
    """
    if schedule_backup():
        config = get_config()
        job = get_job()
        if job:
            return process_job(config, job)
        else:
            logger.error("Didn't get any job from the dispatcher")
            return False
    else:
        logger.error("Failed to schedule a job")
        return False


def schedule_backup():
    """
    Asks dispatcher to schedule a job for this server
    """
    data = {
        "type": "schedule_backup",
        "params": {
        }
    }
    logger.debug("Schedule backup request:")
    logger.debug(data)
    response = get_response(data)
    if response:
        jd = json.JSONDecoder()
        r = jd.decode(response)
        logger.debug(r)
        if r["success"]:
            logger.info("A backup job is successfully registered")
            return True
        else:
            logger.error("Failed to schedule a job: "
                         + jd.decode(decrypt(r["response"]))["error"])
            return False
    else:
        exit_on_error("Empty response from server")


def check_space():
    """
    Checks if it's enough space in TwinDB storage
    Inputs
      config  - backup config
      job     - job order
    Returns always True
    TODO: implement the fucntion
    """
    return True


def check_mysql_permissions(config):
    """
    Checks if MySQL user from backup config has enough privileges to perform backup
    Inputs
      config  - backup config
    Returns
      True    - if all necessary privileges are granted
      False   - if not all necessary privilegers are granted
    """
    try:
        logger.info("Checking if '%s'@'localhost' has enough rights to backup MySQL" % config["mysql_user"])
        con = get_mysql_connection(user=config["mysql_user"], passwd=config["mysql_password"])
        required_priv = ["RELOAD", "LOCK TABLES", "REPLICATION CLIENT", "SUPER", "CREATE TABLESPACE"]
        for priv in required_priv:
            cursor = con.cursor()
            logger.debug("Checking if privilege %s is granted" % priv)
            query = """
            SELECT
                PRIVILEGE_TYPE
            FROM information_schema.USER_PRIVILEGES
            WHERE GRANTEE = '\'%s\'@\\\'localhost\\\'' AND PRIVILEGE_TYPE = %s"""
            logger.debug("Sending query : %s" % query)
            cursor.execute(query, (config["mysql_user"], priv))
            cursor.fetchall()

            if cursor.rowcount > 0:
                logger.info("%20s ... GRANTED" % priv)
                logger.debug("Privilege %s is granted to %s@localhost" % (priv, config["mysql_user"]))
            else:
                logger.info("%20s ... NOT GRANTED" % priv)
                logger.error("Privilege %s is not granted to %s@localhost" % (priv, config["mysql_user"]))
                logger.info("Please execute following SQL to grant %s to user %s@localhost:"
                            % (priv, config["mysql_user"]))
                logger.info("GRANT %s ON *.* TO '%s'@'localhost';" % (priv, config["mysql_user"]))
                return False
            cursor.close()
        con.close()
    except mysql.connector.Error as err:
        logger.error("MySQL Error: %s" % err)
        return False
    logger.debug("MySQL user %s@'localhost' has enough grants to take backup" % config["mysql_user"])
    return True


def take_backup(config, job):
    """
    Meta function that calls actual backup fucntion depending on tool in backup config
    :param config: backup config
    :param job: job order
    :return: what actual backup function returned or -1 if the tool is not supported
    """
    logger.info("Starting backup job")
    try:
        ret = take_backup_xtrabackup(config, job)
    except JobError as err:
        logger.error(err)
        ret = -1
    finally:
        logger.info("Backup job is complete")
    return ret


def start_ssh_cmd(config, job_params, backup_name, stdin, stderr):
    """
    Starts an SSH process to TwinDB storage and saves input in file backup_name
    :param config: backup config
    :param job_params: job parameters
    :param backup_name: file name to save input in
    :param stdin: respective IO handlers
    :param stderr: respective IO handlers
    :return: what SSH process returns
    """
    ssh_process = None
    ssh_cmd = ["ssh", "-oStrictHostKeyChecking=no", "-i", ssh_private_key_file, "-p", str(ssh_port),
               "user_id_%s@%s" % (config["user_id"], job_params["ip"]), "/bin/cat - > %s" % backup_name]
    try:
        logger.debug("Starting SSH process: %r" % ssh_cmd)
        ssh_process = subprocess.Popen(ssh_cmd, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr)
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (ssh_cmd, err))
    return ssh_process


def start_gpg_cmd(stdin, stderr):
    """
    Starts GPG process, encrypts STDIN and outputs in into STDOUT
    :param stdin: respective IO handlers
    :param stderr: respective IO handlers
    :return: what GPG process returns
    """
    gpg_process = None
    gpg_cmd = ["gpg", "--homedir", gpg_homedir, "--encrypt", "--yes", "--batch", "--no-permission-warning",
               "--quiet", "--recipient", server_id]
    try:
        logger.debug("Starting GPG process: %r" % gpg_cmd)
        gpg_process = subprocess.Popen(gpg_cmd, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr)
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (gpg_cmd, err))
    return gpg_process


def take_backup_xtrabackup(config, job):
    """
    # Takes backup copy with XtraBackup
    :param config: backup config
    :param job: job order
    :return: True if backup was successfully taken or False if it has failed
    """
    suffix = "xbstream"
    backup_name = "server_id_%s_%s.%s.gpg" % (server_id, datetime.now().isoformat(), suffix)
    ret_code = 0
    if "params" not in job:
        logger.error("There are no params in the job order")
        return -1
    # Check that job order has all required parameters
    mandatory_params = ["ancestor", "backup_type", "ip", "lsn", "type", "volume_id"]
    for param in mandatory_params:
        if param not in job["params"]:
            logger.error("There is no %s in the job order" % param)
            return -1
    backup_type = job["params"]["backup_type"]
    xtrabackup_cmd = [
        "innobackupex",
        "--stream=xbstream",
        "--user=%s" % config["mysql_user"],
        "--password=%s" % config["mysql_password"],
        "--socket=%s" % get_unix_socket(),
        "--slave-info",
        "--safe-slave-backup",
        "--safe-slave-backup-timeout=3600"]
    if backup_type == 'incremental':
        last_lsn = job["params"]["lsn"]
        xtrabackup_cmd.append("--incremental")
        xtrabackup_cmd.append(".")
        xtrabackup_cmd.append("--incremental-lsn=%s" % last_lsn)
    else:
        xtrabackup_cmd.append(".")
    extra_config = gen_extra_config(config)
    if extra_config:
        xtrabackup_cmd.append("--defaults-extra-file=%s" % extra_config)
    err_descriptors = dict()
    for desc in ["gpg", "ssh", "xtrabackup"]:
        desc_file = ("/tmp/twindb.%s.err" % desc)
        try:
            err_descriptors[desc] = open(desc_file, "w+")
        except IOError as err:
            logger.error("Failed to open file %s. %s" % (desc_file, err))
            return -1
    try:
        logger.debug("Starting XtraBackup process: %r" % xtrabackup_cmd)
        xbk_proc = subprocess.Popen(xtrabackup_cmd, stdout=subprocess.PIPE, stderr=err_descriptors["xtrabackup"])
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (xtrabackup_cmd, err))
        return -1
    gpg_proc = start_gpg_cmd(xbk_proc.stdout, err_descriptors["gpg"])
    ssh_proc = start_ssh_cmd(config, job["params"], backup_name, gpg_proc.stdout, err_descriptors["ssh"])

    try:
        xbk_proc.communicate()
    except IOError as err:
        logger.debug(err)
    try:
        gpg_proc.communicate()
    except IOError as err:
        logger.debug(err)
    try:
        ssh_proc.communicate()
    except IOError as err:
        logger.debug(err)

    ret_code_ssh = ssh_proc.returncode
    ret_code_gpg = gpg_proc.returncode
    ret_code_xbk = xbk_proc.returncode

    err_str = dict()
    for desc in ["gpg", "ssh", "xtrabackup"]:
        err_descriptors[desc].seek(0)
        err_str[desc] = err_descriptors[desc].read()
        if not err_str[desc]:
            err_str[desc] = "no output"

    logger.info("XtraBackup stderr: " + err_str["xtrabackup"])
    logger.info("GPG stderr: " + err_str["gpg"])
    logger.info("SSH stderr: " + err_str["ssh"])

    if ret_code_xbk == 0 and ret_code_gpg == 0 and ret_code_ssh == 0:
        lsn = grep_lsn(err_str["xtrabackup"])
        if not lsn:
            logger.error("Could not find LSN in XtrabBackup output")
            return -1
        file_size = get_backup_size(config, job["params"], backup_name)
        if not file_size:
            logger.error("Backup copy size must not be zero")
            return -1
        volume_id = job["params"]["volume_id"]
        ancestor = job["params"]["ancestor"]
        if not record_backup(backup_name, volume_id, file_size, lsn, ancestor):
            logger.error("Failed to save backup copy details")
            return -1
    else:
        if ret_code_xbk != 0:
            logger.error("XtraBackup exited with code %d" % ret_code_xbk)
        if ret_code_gpg != 0:
            logger.error("GPG exited with code %d" % ret_code_gpg)
        if ret_code_ssh != 0:
            logger.error("SSH exited with code %d" % ret_code_ssh)
        logger.error("Failed to take backup")
        return -1
    for f in [extra_config, "/tmp/twindb.xtrabackup.err", "/tmp/twindb.gpg.err", "/tmp/twindb.ssh.err"]:
        if os.path.isfile(f):
            try:
                os.remove(f)
            except IOError as err:
                logger.error("Failed to remove file %s. %s" % (f, err))
    return ret_code


def gen_extra_config(config):
    """
    Generates MySQL config with datadir option
    :param config: backup config
    :return: File name with MySQL config or None if error happened
    """
    try:
        f, extra_config = tempfile.mkstemp()
        os.write(f, "[mysqld]\n")
        con = get_mysql_connection(config["mysql_user"], config["mysql_password"])
        cur = con.cursor()
        cur.execute("SELECT @@datadir")
        row = cur.fetchone()
        os.write(f, 'datadir="%s"\n' % row[0])
        cur.close()
        os.close(f)
    except IOError as err:
        logger.error("Failed to generate extra defaults file. %s" % err)
        extra_config = None
    return extra_config


def get_backup_size(config, job_params, backup_name):
    """
    Logs in to TwinDB storage and get size of backup
    :param config: backup config
    :param job_params: job parameters
    :param backup_name: file name with backup
    :return: size of backup in bytes or zeor if error happened
    """
    logger.debug("Getting size of %s" % backup_name)
    ssh_cmd = ["ssh", "-oStrictHostKeyChecking=no", "-i", ssh_private_key_file, "-p", str(ssh_port),
               "user_id_%s@%s" % (config["user_id"], job_params["ip"]), "/bin/du -b %s" % backup_name]
    try:
        process = subprocess.Popen(ssh_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        cout, cerr = process.communicate()

        if process.returncode != 0:
            raise subprocess.CalledProcessError(process.returncode, ' '.join(ssh_cmd), cout)

        cout_lines = cout.split()
        if len(cout_lines) < 1:
            raise subprocess.CalledProcessError(process.returncode, ' '.join(ssh_cmd), cout)

        backup_size = int(cout_lines[0])
    except subprocess.CalledProcessError as e:
        logger.error("Failed to get size of backup %s" % backup_name)
        logger.error(str(e))
        return 0
    except OSError as e:
        logger.error("Failed to get size of backup %s" % backup_name)
        logger.error("Failed to run command %r: %s" % (ssh_cmd, e))
        return 0
    logger.debug("Size of %s = %d bytes (%s)" % (backup_name, backup_size, h_size(backup_size)))
    return backup_size


def handler_send_key(job):
    """
    Processes send_key job
    :param job: job order
    :return: nothing
    """
    # Get owner of the GPG key
    cmd_1 = ["gpg", "--list-packets"]
    try:
        gpg_pub_key = job["params"]["gpg_pub_key"]
        if gpg_pub_key:
            logger.debug("Starting %r" % cmd_1)
            p1 = subprocess.Popen(cmd_1, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            cout, cerr = p1.communicate(gpg_pub_key)
            keyid = "Unknown"
            for line in cout.split("\n"):
                if "keyid:" in line:
                    keyid = line.replace("keyid:", "").strip()
                    break
            logger.debug("Requestor's public key id is %s" % keyid)
        else:
            logger.error("Requestor public key is empty")
            return -1
    except OSError as err:
        logger.error("Failed to run command %r: %s" % (cmd_1, err))
        return -1
    # Import public GPG key. It's a user public key sent by the dispatcher
    try:
        logger.debug("Importing requestor's key %s" % keyid)
        cmd_1 = ["gpg", "--import"]
        p1 = subprocess.Popen(cmd_1, stdin=subprocess.PIPE)
        p1.communicate(gpg_pub_key)
    except OSError as err:
        logger.error("Failed to run command %r: %s" % (cmd_1, err))
        return -1
    # Get private key and encrypt it
    gpg_pub_key = job["params"]["gpg_pub_key"]
    if gpg_pub_key:
        logger.debug("Exporting private key of server %s" % server_id)
        cmd_1 = ["gpg", "--armor", "--export-secret-key", server_id]
        cmd_2 = ["gpg", "--armor", "--encrypt", "--sign", "--batch", "-r", keyid, "--local-user", server_id,
                 "--trust-model", "always"]
        try:
            logger.debug("Starting %r" % cmd_1)
            p1 = subprocess.Popen(cmd_1, stdout=subprocess.PIPE)
        except OSError as err:
            logger.error("Failed to run command %r: %s" % (cmd_1, err))
            return -1
        try:
            logger.debug("Starting %r" % cmd_2)
            p2 = subprocess.Popen(cmd_2, stdin=p1.stdout, stdout=subprocess.PIPE)
            cout, cerr = p2.communicate()
            enc_private_key = cout
            logger.debug("Encrypted private key %s" % enc_private_key)
        except OSError as err:
            logger.error("Failed to run command %r: %s" % (cmd_2, err))
            return -1
        # Now send the private key to dispatcher
        data = {
            "type": "send_key",
            "params": {
                "enc_private_key": enc_private_key,
                "job_id": job_id
            }
        }
        get_response(data)
    else:
        logger.error("The job order requested send_key, but no public key was provided")
        return -1


def is_dir_empty(directory):
    """
    Checks if directory is empty
    :param directory: directory name
    :return: True if the directory is empty of False otherwise
    """
    return len(os.listdir(directory)) == 0


def restore_backup(config, job):
    """
    Meta function that calls actual restore fucntion depending on tool in backup config
    :param config: backup config
    :param job: job order
    :return: what actual restore function returned or -1 if the tool is not supported
    """
    logger.info("Starting restore job: %s"
                % json.dumps(job, indent=4, sort_keys=True))
    try:
        ret = restore_xtrabackup(config, job)
    except JobError as err:
        logger.error(err)
        return -1
    finally:
        logger.info("Restore job is complete")
    return ret


def restore_xtrabackup(config, job):
    """
    # Restores backup copy with XtraBackup
    # Inputs
    #   config  - backup config
    #   job     - job order
    # Returns
    #   True    - if backup successfully restored
    #   False   - if restore job failed
    :param config: backup config received from the dispatcher
    :param job: job order
    :return:
    """
    if "params" not in job:
        logger.error("There are no params in the job order")
        return -1
    # Check that job order has all required parameters
    mandatory_params = ["backup_copy_id", "restore_dir", "server_id"]
    for param in mandatory_params:
        if param not in job["params"]:
            logger.error("There is no %s in the job order" % param)
            return -1
    dst_dir = job["params"]["restore_dir"]
    try:
        if os.path.isdir(dst_dir):
            if is_dir_empty(dst_dir):
                logger.info("Directory %s exists. But it's empty, so we can restore backup here" % dst_dir)
            else:
                logger.error("Directory %s exists and isn't empty, so we can not restore backup here" % dst_dir)
                return -1
        else:
            os.makedirs(dst_dir)
    except IOError as err:
        logger.error(err)
        logger.error("Can't use directory %s as destination for backup" % dst_dir)
        return -1

    try:
        full_copy = True
        backups_chain = get_backups_chain(job["params"]["backup_copy_id"])
        if not backups_chain:
            raise JobError("Failed to get backups chain from dispatcher")
        for backup_copy in backups_chain:
            logger.debug("Processing backup copy %r" % backup_copy)

            if full_copy:
                if not backup_copy["full"]:
                    raise JobError("Expected full copy, but it's not")
                if not extract_archive(config, backup_copy, dst_dir):
                    raise JobError("Failed to extract %s" % backup_copy["name"])
                # We restored full backup copy in dst_dir
                try:
                    xb_err = open('/tmp/twindb.xb.err', "w+")
                except IOError as err:
                    raise JobError("Failed to open /tmp/twindb.xb.err: %s" % err)
                # if this is the last copy in the chain then --apply-log
                # otherwise just apply the redo log
                if backup_copy["backup_copy_id"] == job["params"]["backup_copy_id"]:
                    xb_cmd = ["innobackupex", "--apply-log", dst_dir]
                else:
                    xb_cmd = ["innobackupex", "--apply-log", "--redo-only", dst_dir]
                try:
                    p_xb = subprocess.Popen(xb_cmd, stdout=xb_err, stderr=xb_err)
                    p_xb.communicate()
                except OSError as err:
                    raise JobError("Failed to run %r: %s" % (xb_cmd, err))
                finally:
                    xb_err.seek(0)
                    logger.info("innobackupex stderr: " + xb_err.read())
                    os.remove('/tmp/twindb.xb.err')
                if p_xb.returncode != 0:
                    raise JobError("Failed to apply log on full copy %s" % backup_copy["name"])
                full_copy = False
            else:
                inc_dir = tempfile.mkdtemp()
                if not extract_archive(config, backup_copy, inc_dir):
                    raise JobError("Failed to extract %s in %s" % (backup_copy["name"], inc_dir))
                try:
                    xb_err = open("/tmp/twindb.xb.err", "w+")
                except IOError as err:
                    raise JobError("Failed to open /tmp/twindb.xb.err: %s" % err)
                xb_cmd = ["innobackupex", "--apply-log"]
                if backup_copy["backup_copy_id"] != job["params"]["backup_copy_id"]:
                    xb_cmd.append("--redo-only")
                xb_cmd.append('--incremental-dir=%s' % inc_dir)
                xb_cmd.append(dst_dir)
                try:
                    p_xb = subprocess.Popen(xb_cmd, stdout=xb_err, stderr=xb_err)
                    p_xb.wait()
                except OSError as err:
                    raise JobError("Failed to run %r: %s" % (xb_cmd, err))
                finally:
                    xb_err.seek(0)
                    logger.info("innobackupex stderr: " + xb_err.read())
                    os.remove('/tmp/twindb.xb.err')
                if p_xb.returncode != 0:
                    raise JobError("Failed to apply log on copy %s" % backup_copy["name"])
                shutil.rmtree(inc_dir)
            logger.info("Successfully restored backup %s in %s" % (backup_copy["name"], dst_dir))
    except JobError as err:
        logger.error(err)
        logger.error("Failed to restore backup")
        return -1
    finally:
        for f in ["/tmp/twindb.xb.err", "/tmp/twindb.gpg.err", "/tmp/twindb.ssh.err"]:
            if os.path.isfile(f):
                os.remove(f)
    return 0


def extract_archive(config, arc, dst_dir):
    """
    Extracts an Xtrabackup archive arc in  dst_dir
    :param config: backup config
    :param arc: dictionary with archive to extract
    :param dst_dir: local destination directory
    :return:    True - if archive is successfully extracted.
                False - if error happened
    """
    mandatory_params = ["backup_copy_id", "name", "ip"]
    for param in mandatory_params:
        if param not in arc:
            logger.error("There is no %s in the archive parameters" % param)
            return False
    logger.info("Extracting %s in %s" % (arc["name"], dst_dir))
    ssh_cmd = [
        "ssh",
        "-oStrictHostKeyChecking=no",
        "-i", ssh_private_key_file,
        "-p", str(ssh_port),
        "user_id_%s@%s" % (config["user_id"], arc["ip"]),
        "/bin/cat %s" % arc["name"]
    ]
    gpg_cmd = ["gpg", "--decrypt"]
    xb_cmd = ["xbstream", "-x"]

    desc_file = None
    try:
        err_desc = dict()
        for desc in ["xb", "gpg", "ssh"]:
            desc_file = ("/tmp/twindb.%s.err" % desc)
            err_desc[desc] = open(desc_file, "w+")
    except IOError as err:
        logger.error("Failed to open %s: %s" % (desc_file, err))
        return False

    logger.info("Starting: %r" % ssh_cmd)
    p1 = subprocess.Popen(ssh_cmd, stdout=subprocess.PIPE, stderr=err_desc["ssh"])
    logger.info("Starting: %r" % gpg_cmd)
    p2 = subprocess.Popen(gpg_cmd, stdin=p1.stdout, stdout=subprocess.PIPE,
                          stderr=err_desc["gpg"])
    logger.info("Starting: %r" % xb_cmd)
    p3 = subprocess.Popen(xb_cmd, stdin=p2.stdout, stdout=subprocess.PIPE,
                          stderr=err_desc["xb"], cwd=dst_dir)
    p1.wait()
    p2.wait()
    p3.wait()
    for desc in ["xb", "gpg", "ssh"]:
        err_desc[desc].seek(0)

    logger.info("SSH stderr: " + err_desc["ssh"].read())
    logger.info("GPG stderr: " + err_desc["gpg"].read())
    logger.info("xbstream stderr: " + err_desc["xb"].read())
    if p1.returncode != 0 or p2.returncode != 0 or p3.returncode != 0:
        logger.info("Failed to extract backup %s in %s" % (arc["name"], dst_dir))
        return False
    desc_file = None
    try:
        for desc in ["xb", "gpg", "ssh"]:
            desc_file = ("/tmp/twindb.%s.err" % desc)
            if os.path.isfile(desc_file):
                os.remove(desc_file)
    except IOError as err:
        logger.error("Failed to open %s: %s" % (desc_file, err))
    logger.info("Extracted successfully %s in %s" % (arc["name"], dst_dir))
    return True


def get_backups_chain(backup_copy_id):
    """
    Gets a chain of parents of the given backup_copy_id
    :param backup_copy_id: backup_copy_id
    :return: list of dictionaries with backup copy params:
        {
        "backup_copy_id": 188,
        "name": "server_id_479a41b3-d22d-41a8-b7d3-4e40302622f6_2015-04-06T15:10:46.050984.tar.gp",
        "ip": "127.0.0.1",
        full: True
        },
        {...}
    """
    logger.debug("Getting backups chain for backup_copy_id %d" % backup_copy_id)
    response_body = "Empty response"
    backups_chain = None
    try:
        data = {
            "type": "get_backups_chain",
            "params": {
                "backup_copy_id": backup_copy_id
            }
        }
        response_body = get_response(data)
        if not response_body:
            logger.debug("Empty response from dispatcher")
            return None
        d = json.JSONDecoder()
        response_body_decoded = d.decode(response_body)
        if response_body_decoded:
            msg_decrypted = decrypt(response_body_decoded["response"])
            msg_pt = d.decode(msg_decrypted)
            backups_chain = msg_pt["data"]
            logger.info("Got backups chain:\n%s"
                        % json.dumps(backups_chain, indent=4, sort_keys=True))
            if msg_pt["error"]:
                logger.error(msg_pt["error"])
    except exceptions.KeyError as err:
        logger.error("Failed to decode %s" % response_body)
        logger.error(err)
        return None
    return backups_chain


def grep_lsn(output):
    """
    Finds LSN in XtraBackup output
    :param output: string with Xtrabackup output
    :return: LSN
    """
    lsn = None
    for line in output.split("\n"):
        if line.startswith("xtrabackup: The latest check point (for incremental):"):
            lsn = line.split("'")[1]
    return lsn


def get_unix_socket():
    """
    Finds MySQL socket
    :return: path to unix socket or None if not found
    """
    cmd = ["lsof", "-U", "-c", "/^mysqld$/", "-a", "-F", "n"]
    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        cout, cerr = p.communicate()
        # Outputs socket in format
        # # lsof -U -c mysqld -a -F n
        # p11029
        # n/var/lib/mysql/mysql.sock
        mysql_socket = cout.split()[1][1:]
        if not os.path.exists(mysql_socket):
            return None
    except OSError as err:
        logger.error("Failed to run command %r. %s" % (cmd, err))
        return None
    return mysql_socket


def pid_exists(pid):
    """
    Checks whether pid exists in the current process table.
    """
    if pid < 0:
        return False
    try:
        os.kill(pid, 0)
    except OSError as e:
        return e.errno == errno.EPERM
    else:
        return True


class TimeoutExpired(Exception):
    pass


def wait_pid(pid, timeout=None):
    """Wait for process with pid 'pid' to terminate and return its
    exit status code as an integer.

    If pid is not a children of os.getpid() (current process) just
    waits until the process disappears and return None.

    If pid does not exist at all return None immediately.

    Raise TimeoutExpired on timeout expired (if specified).
    """

    def check_timeout(timeout_delay):
        if timeout is not None:
            if time.time() >= stop_at:
                raise TimeoutExpired
        time.sleep(timeout_delay)
        return min(timeout_delay * 2, 0.04)

    if timeout is not None:
        waitcall = lambda: os.waitpid(pid, os.WNOHANG)
        stop_at = time.time() + timeout
    else:
        waitcall = lambda: os.waitpid(pid, 0)

    delay = 0.0001
    while 1:
        try:
            retpid, status = waitcall()
        except OSError, err:
            if err.errno == errno.EINTR:
                delay = check_timeout(delay)
                continue
            elif err.errno == errno.ECHILD:
                # This has two meanings:
                # - pid is not a child of os.getpid() in which case
                #   we keep polling until it's gone
                # - pid never existed in the first place
                # In both cases we'll eventually return None as we
                # can't determine its exit status code.
                while 1:
                    if pid_exists(pid):
                        delay = check_timeout(delay)
                    else:
                        return
            else:
                raise
        else:
            if retpid == 0:
                # WNOHANG was used, pid is still running
                delay = check_timeout(delay)
                continue
            # process exited due to a signal; return the integer of
            # that signal
            if os.WIFSIGNALED(status):
                return os.WTERMSIG(status)
            # process exited using exit(2) system call; return the
            # integer exit(2) system call has been called with
            elif os.WIFEXITED(status):
                return os.WEXITSTATUS(status)
            else:
                # should never happen
                raise RuntimeError("unknown process exit status")


def remove_pid():
    """
    Removes pid file
    :return: nothing
    Exits if error happened
    """
    if os.path.isfile(pid_file):
        try:
            os.remove(pid_file)
        except IOError as err:
            exit_on_error("Failed to remove file %s. %s" % (pid_file, err))


def check_pid():
    """
    Checks if pid file already exists.
    If it does it detects whether twindb agent is already running.
    If the pid file is stale it removes it.
    :return: True if pid file doesn't exist or was stale and it was removed.
    False if twindb agent is running or error happened
    """
    if os.path.isfile(pid_file):
        pid = read_pid()
        if pid_exists(pid):
            try:
                f = open("/proc/%d/cmdline" % pid, "r")
                cmdline = f.readline()
                f.close()
                if "twindb.py" in cmdline:
                    # The process is a live twindb agent
                    return False
                else:
                    # It's some other process, not a twindb agent
                    remove_pid()
            except IOError as err:
                logger.error(err)
                remove_pid()
        else:
            remove_pid()

    # pid file doesn't exist
    return True


def read_pid():
    """
    Read pid from pid_file
    :return: pid or zero if pid file doesn't exist
    """
    try:
        f = open(pid_file, 'r')
        pid = int(f.readline())
        f.close()
    except IOError as err:
        logger.error("Couldn't read from %s: %s" % (pid_file, err))
        return 0
    return int(pid)


def write_pid():
    """
    Writes pid of the current process to the pid file
    :return: nothing.
    Exists if error happened
    """
    try:
        f = open(pid_file, "w")
        f.write(str(os.getpid()))
        f.close()
    except IOError as err:
        logger.error("Couldn't save process id in " + pid_file)
        exit_on_error(err)


def cleanup(signum, frame):
    """
    Cleans up when TwinDB agent exists
    :param signum:
    :param frame:
    :return:
    """

    logger.info("Cleaning up on signal " + str(signum))
    logger.debug("Frame %r" % frame)
    logger.info("TwinDB agent is ready to exit")
    sys.exit(0)


def exit_on_error(message):
    """
    Reports error removes pid and exits
    :rtype : object
    :param message: message to display
    :return:
    """
    logger.error(message)
    logger.debug(traceback.format_exc())
    sys.exit(2)


# Stops TwinDB agent

def stop():
    """
    Stops TwinDB agent
    :return: nothing
    """

    logger.info("Shutting down TwinDB agent")

    if not os.path.exists(pid_file):
        logger.info("Pid file %s does not exist. Probably twindb agent isn't running" % pid_file)
        sys.exit(0)
    pid = None

    try:
        pid = read_pid()

        if pid_exists(pid):
            os.kill(pid, signal.SIGTERM)
            wait_pid(pid, 300)

        remove_pid()
    except OSError as err:
        logger.error("Couldn't kill process %d" % pid)
        exit_on_error(err)
    except IOError as err:
        logger.error("Couldn't read from %s" % pid_file)
        exit_on_error(err)

    logger.info("TwinDB agent successfully shut down")
    sys.exit(0)


def start():
    """
    Starts TwinDB agent
    :return: nothing
    """
    global job_id
    logger.info("Starting TwinDB agent")

    try:
        while True:
            read_config()
            config = get_config()
            if not config:
                time.sleep(check_period)
                continue
            report_sss(config)
            report_agent_privileges(config)
            job = get_job()
            if job:
                process_job(config, job)
            time.sleep(check_period)
    except SystemExit:
        remove_pid()
    sys.exit(0)


def process_job(config, job):
    """
    Processes job
    :param config: backup config
    :param job: job order
    :return: what respective job function returns or False if error happens
    """
    global job_id
    # Check to see that the twindb_agent MySQL user has enough privileges
    username = config["mysql_user"]
    password = config["mysql_password"]

    job_id = int(job["job_id"])

    try:
        mysql_access_available, missing_mysql_privileges = has_mysql_access(username, password, grant_capability=False)
        if not mysql_access_available:
            logger.error("The MySQL user %s does not have all the required privileges." % username)
            if missing_mysql_privileges:
                raise JobError("You can grant the required privileges by executing the following SQL: "
                               "GRANT %s ON *.* TO '%s'@'localhost' IDENTIFIED BY '%s'; FLUSH PRIVILEGES;"
                               % (','.join(missing_mysql_privileges), username, password))
        if not job["start_scheduled"]:
            raise JobError("Job start time isn't set")
        start_scheduled = int(job["start_scheduled"])
        now = int(time.time())
        if now < start_scheduled:
            raise JobTooSoonError("Job is scheduled on %s, now %s"
                                  % (time.ctime(start_scheduled), time.ctime(now)))
        logger.info("Processing job_id = %d", int(job_id))
        notify_params = {"event": "start_job", "job_id": int(job_id)}
        if not log_job_notify(notify_params):
            raise JobError("Failed to notify dispatcher abot job start")
        ret = -1
        if job["type"] == "backup":
            ret = take_backup(config, job)
        elif job["type"] == "restore":
            ret = restore_backup(config, job)
        elif job["type"] == "send_key":
            ret = handler_send_key(job)
        else:
            raise JobError("Unsupported job type " + job["type"])
        notify_params = {"event": "stop_job", "job_id": job["job_id"], "ret_code": ret}
        log_job_notify(notify_params)
    except JobError as err:
        logger.error("Job error: %s", err)
        notify_params = {"event": "stop_job", "job_id": job["job_id"], "ret_code": -1}
        log_job_notify(notify_params)
        return False
    except JobTooSoonError as err:
        logger.debug(err)
        return False
    finally:
        job_id = 0
    return True


def setup_logging():
    """
    Setups logging handlers, formats
    :return:
    """
    global logger

    logdir = "/var/log/twindb"
    try:
        if not os.path.exists(logdir):
            os.mkdir(logdir)
    except OSError as err:
        print("Failed to create directory %s: %s" % (logdir, err))
        exit(2)

    file_handler = logging.handlers.WatchedFileHandler("%s/twindb-agent.log" % logdir)
    console_handler = logging.StreamHandler()
    remote_handler = RlogHandler()

    fmt_str = '%(name)s: %(levelname)s: %(funcName)s():%(lineno)d: %(message)s'
    remote_format = logging.Formatter(fmt_str)
    console_format = logging.Formatter("%(asctime)s: " + fmt_str)

    file_handler.setFormatter(console_format)
    console_handler.setFormatter(console_format)
    remote_handler.setFormatter(remote_format)

    logger.addHandler(file_handler)
    logger.addHandler(console_handler)
    logger.addHandler(remote_handler)
    logger.setLevel(logging.INFO)


def main():
    """
    Main function
    Parses options, creates log class etc
    :return:
    """
    global server_id
    global check_period
    global host
    global mysql_user
    global mysql_password
    global debug
    global logger

    setup_logging()
    read_config()
    # Signal handlers
    signal.signal(signal.SIGTERM, cleanup)
    signal.signal(signal.SIGINT, cleanup)

    # Not clear why I ignore SIGCHLD.
    # it make subprocess.call throw exception, so I comment it out so far
    # signal.signal(signal.SIGCHLD, signal.SIG_IGN)

    opts = []
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'd:i:ghu:p:',
                                   ["start", "stop", "help", "register=", "dispatcher=", "user=", "password=",
                                    "version",
                                    "unregister", "delete-backups", "backup", "is-registered"])
    except getopt.GetoptError as err:
        print(err)
        usage()
        sys.exit(2)
    # Set options first
    action = None
    delete_backups = False
    regcode = None
    for opt, arg in opts:
        if opt == '-i':
            check_period = int(arg)
        elif opt in ['-d', '--dispatcher']:
            host = arg
        elif opt in ['-u', '--user']:
            mysql_user = arg
        elif opt in ['-p', '--password']:
            mysql_password = arg
        elif opt == '-g':
            debug = True
            logger.setLevel(logging.DEBUG)
        elif opt == '--start':
            action = "start"
        elif opt == '--stop':
            action = "stop"
        elif opt == '--register':
            action = "register"
            regcode = arg
        elif opt == '--unregister':
            action = "unregister"
        elif opt == '--delete-backups':
            delete_backups = True
        elif opt == '--backup':
            action = "backup"
        elif opt == '--is-registered':
            action = "is-registered"
        elif opt in ["-h", "--help"]:
            usage()
            sys.exit(0)
        elif opt in ["--version"]:
            print("%s" % agent_version)
            sys.exit(0)
        else:
            usage()
            sys.exit(2)

    # Check that the option combinations have been used correctly
    if delete_backups and action != "unregister":
        logger.error("Hey, --delete-backups can be used only with --unregister")
        sys.exit(2)

    # before we do *anything* we must ensure server_id is generated or read from config
    # Read or generate server id
    if not server_id:
        server_id = str(uuid.uuid4())
        save_config()

    # At this point server_id must be set
    if not server_id:
        print("Error: Server id is empty. Can not continue operation", file=sys.stderr)
        sys.exit(2)

    # Check the environment and set it up correctly if its not already set
    check_env()

    if check_period < 1:
        check_period = 1

    if action == "start":
        # Create the PID if it does not already exist
        # If the PID already exists then we bail out
        if check_pid():
            write_pid()
        else:
            exit_on_error("Another instance of TwinDB agent is running. Exiting")

        while True:
            # It'll keep checking until the agent is registered
            # Then it starts
            if is_registered():
                start()
            else:
                logger.error("The server must be registered first. Run following command:\n\n"
                             "twindb --register <registration code>\n")
                logger.error("Get your code on https://console.twindb.com/?get_code")
            time.sleep(check_period)
    elif action == "stop":
        stop()
    elif action == "register":
        action_handler_register(regcode)
    elif action == "unregister":
        if not is_registered():
            logger.error("The server must be registered first. Run following command:"
                         "\n\ntwindb --register <registration code>\n")
            logger.error("Get your code on https://console.twindb.com/?get_code")
            sys.exit(2)
        if action_handler_unregister(delete_backups):
            stop()
    elif action == "backup":
        if not is_registered():
            logger.error("The server must be registered first. Run following command:"
                         "\n\ntwindb --register <registration code>\n")
            logger.error("Get your code on https://console.twindb.com/?get_code")
            sys.exit(2)
        if not action_handler_backup():
            logger.error("Failed to take backup")
            sys.exit(2)
    elif action == "is-registered":
        if is_registered():
            print("YES")
        else:
            print("NO")
        sys.exit(0)
    else:
        # If we got there print usage() and exit
        logger.error("Failed to parse command line options")
        usage()
        sys.exit(2)


if __name__ == "__main__":
    logger = logging.getLogger('twindb')
    main()
